/*
 This file is part of the digitalSTROM ds485-core library.

 Copyright (C) 2010 digitalSTROM.org, Schlieren, Switzerland

 digitalSTROM ds485-core library is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation version 2
 of the License.

 digitalSTROM ds485-core library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with digitalSTROM ds485-core library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#ifndef DS485QP_H_
#define DS485QP_H_

#include "digitalSTROM/ds.h"
#include "qp_port.h"
#include "digitalSTROM/ds485.h"

#ifndef FALSE
# define FALSE 0
#endif
#ifndef TRUE
# define TRUE 1
#endif

#ifndef bool
#define bool int
#endif

/*
 * Serial Port Settings
 */
#define dS485_Baudrate 115200
#define dS485_Databits 8
#define dS485_Stopbits 1
#define dS485_Parity 0


/**
 * Get destination address from DS485MessageEvt
 */
#define DS485_GetDestination(evt_)  ((evt_)->message[DS485_DEST_OFFSET] >> 2)

/**
 * Get source address from DS485MessageEvt
 */
#define DS485_GetSource(evt_)       ((evt_)->message[DS485_SRC_OFFSET] >> 2)

/**
 * Get command from DS485MessageEvt
 */
#define DS485_GetCommand(evt_)      ((evt_)->message[DS485_CL_OFFSET] >> 4 )

/**
 * Get payload length from DS485MessageEvt
 */
#define DS485_GetPayloadLength(evt_) ((evt_)->message[DS485_CL_OFFSET] & 0x0F )

/**
 * Set source address to DS485MessageEvt
 */
#define DS485_SetSource(evt_, src_) do { \
    (evt_)->message[DS485_SRC_OFFSET] &= 0x03; \
    (evt_)->message[DS485_SRC_OFFSET] |= (src_) << 2; \
} while(0);

/**
 * Returns true if destination address is broadcast
 */
#define DS485_IsBroadcast(evt_)     (((evt_)->message[DS485_DEST_OFFSET] & DS485_BROADCAST_MASK) != 0)

/**
 * Returns true if long message
 */
#define DS485_IsLongMessage(evt_)   (((evt_)->message[DS485_CL_OFFSET] & DS485_LENGTH_MASK) == 0)

/**
 * Returns true if message is a request
 */
#define DS485_IsRequest(evt_)       (DS485_GetCommand(evt_) == DS485_REQUEST)

/**
 * Returns true if message is a response
 */
#define DS485_IsResponse(evt_)      (DS485_GetCommand(evt_) == DS485_RESPONSE)

/**
 * Returns true if message is a request
 */
#define DS485_IsEvent(evt_)       (DS485_GetCommand(evt_) == DS485_EVENT)

/**
 * Returns the message ID (dSM-API).
 */
#define DS485_GetMessageId(evt_)    ((!DS485_IsLongMessage(evt_)) ? ((evt_)->message[DS485_DATA_OFFSET]) : ((evt_)->message[DS485_DATA_OFFSET+1]))

/**
 * Copy 16-bit value from byte buffer least significant byte first.
 */
#define DS485_CopyBufToValue16(buf8_, val16_)   memcpy(&(val16_), (buf8_), 2)

/**
 * Copy 32-bit value from byte buffer least significant byte first.
 */
#define DS485_CopyBufToValue32(buf8_, val32_)   memcpy(&(val32_), (buf8_), 4)

/**
 * Copy 16-bit value to byte buffer least significant byte first.
 */
#define DS485_CopyValue16ToBuf(buf8_, val16_)   memcpy((buf8_), &(val16_), 2)

/**
 * Copy 32-bit value to byte buffer least significant byte first.
 */
#define DS485_CopyValue32ToBuf(buf8_, val32_)   memcpy((buf8_), &(val32_), 4)

/**
 * Returns parameter from request/response
 * Note: Byte order is least significant byte first!
 */
#define DS485_GetParameter(evt_, index_) \
    ((((uint16_t)((evt_)->message[DS485_DATA_OFFSET+2+2*(index_)])) << 8) | \
                 ((evt_)->message[DS485_DATA_OFFSET+1+2*(index_)]))

/**
 * Set parameter to request/response
 * Note: Byte order is least significant byte first!
 */
#define DS485_SetParameter(evt_, index_, par_) do { \
    (evt_)->message[DS485_DATA_OFFSET+1+2*(index_)] = (par_) >> 8; \
    (evt_)->message[DS485_DATA_OFFSET+2+2*(index_)] = (par_); \
} while (0)

/**
 * Set message ID to request/response
 */
#define DS485_SetMsgID(evt_, msgID_) do { \
    (evt_)->message[DS485_DATA_OFFSET] = (msgID_); \
} while (0)

/**
 * Insert dsid into message
 */
#define DS485_SetDSID(evt_, dsid_) do { \
    memcpy(&evt_->message[DS485_DATA_OFFSET], &dsid_, sizeof(me->dsmid)); \
} while (0)

/**
 * Set parameter to long request/response
 * Note: Byte order is least significant byte first!
 */
#define DS485_SetLongParameter(evt_, index_, par_) do { \
    (evt_)->message[DS485_DATA_OFFSET+2+2*(index_)] = (par_) >> 8; \
    (evt_)->message[DS485_DATA_OFFSET+3+2*(index_)] = (par_); \
} while (0)

/**
 * Set message ID to long request/response
 */
#define DS485_SetLongMsgID(evt_, msgID_) do { \
    (evt_)->message[DS485_DATA_OFFSET+1] = (msgID_); \
} while (0)

/**
 * Set message destination address
 */
#define DS485_SetDestination(evt_, dest_)  ((evt_)->message[DS485_DEST_OFFSET] |= (dest_ << 2))

/**
 * DS485 configuration event structure
 */
typedef struct DS485ConfigEvtTag {
    QEvent super;               /* derives from QEvent */
    uint32_t id;                  /* device dsid */
    uint8_t reqAddress;         /* ds485 default address */
} DS485ConfigEvt;

/**
 * DS485 address event structure
 */
typedef struct DS485AddressEvtTag {
    QEvent super;               /* derives from QEvent */
    uint8_t deviceAddress;      /* dS485 device address */
} DS485AddressEvt;

/**
 * DS485 standard message event structure
 */
typedef struct {
    QEvent super;                               /* derives from QEvent */
    uint16_t size;                              /* message size (including CRC-16) */
    uint8_t message[DS485_SHORT_MESSAGE_SIZE];  /* message data (including CRC-16) */
} DS485ControlMessageEvt;

/**
 * DS485 message event structure
 */
typedef struct {
    QEvent super;                               /* derives from QEvent */
    uint16_t size;                              /* message size (including CRC-16) */
    uint8_t message[DS485_SHORT_MESSAGE_SIZE];  /* message data (including CRC-16) */
#ifdef __linux__
    dsid_t destinationAddr;
    dsid_t sourceAddr;
#endif
} DS485MessageEvt;

/**
 * DS485 message event structure
 */
typedef struct {
    QEvent super;                               /* derives from QEvent */
    uint16_t size;                              /* message size (including CRC-16) */
    uint8_t message[DS485_LONG_MESSAGE_SIZE];  /* message data (including CRC-16) */
#ifdef __linux__
    dsid_t destinationAddr;
    dsid_t sourceAddr;
#endif
} DS485LongMessageEvt;

#ifdef __linux__
typedef struct {
    QEvent super;                               /* derives from QEvent */
    int error;
} DS485ResponseEvt;
#endif

/**
 * DS485 statistics request event
 */
typedef struct {
    QEvent super;
    QActive * requester;
} DS485ReqStatisticsEvt;

/**
 * DS485 statistics response event
 */
typedef struct {
    QEvent super;
    DS485Statistics statistics;
} DS485ResStatisticsEvt;

#ifdef __linux__

typedef struct {
    QEvent super;
    dsid_t  dsid[DS485_MAX_SLAVE_ADDRESS+1];
} DS485_DsidResponseEvt;
#endif

/**
 * Data model publish config event
 */
typedef struct {
    QEvent super;
    dsid_t dsmid;
} DatamodelConfigEvt;


/**
 * DS485 AO constructor
 */
void DS485_ctor(void);

/**
 * Create DS485 message
 */
DS485MessageEvt * DS485_CreateMessage(uint8_t dest, uint8_t src, uint8_t msgCount, uint8_t cmd, uint8_t payloadLength);

/*
 * Map new protocol data event to short message event
 */
#define DS485DataMessageEvt DS485MessageEvt

/**
 * Create DS485 data message
 */
DS485MessageEvt * DS485_CreateDataMessage(const dsid_t *dest, DS485CommandType type, uint8_t payloadLength);

/**
 * Opaque pointer to DS485 AO
 */
extern QActive * const AO_DS485;

/**
 * DS485 message queue
 */
extern QEQueue ds485TxMessageQueue;

extern QActive * const AO_Receiver;
extern QActive * const AO_DS485;
extern QActive * const AO_DsmApi;

void Receiver_ctor(void);
void DS485_ctor(void);
void DSMApi_ctor(void);

#endif /* DS485QP_H_ */
